from ...utilities import OPTIONS

from .cage import *
from .macro_molecule import *
from .molecule import *
from .periodic import *
from .polymer import *
from .struct_unit import *
from .macrocycle import *
from .rotaxane import *

# Toggle the caching of molecules.
OPTIONS['cache'] = True
